{- | Check a policy against a usage. -}
module Main where

import System.Environment
import qualified Text.ParserCombinators.Parsec as P
import Control.Monad
import qualified Data.Map as M
import qualified Data.Set as S
import Data.List
import Data.Char

import Common
import CFG
import Policy
import Usage
import Parser

-- import Debug.Trace

-- * Set utilities

unionsS :: Ord a => S.Set (S.Set a) -> S.Set a
unionsS = S.unions . S.toList

concatMapS :: (Ord a, Ord b) => (a -> S.Set b) -> S.Set a -> S.Set b
concatMapS f = unionsS . S.map f

bindS :: (Ord a, Ord b) => S.Set a -> (a -> S.Set b) -> S.Set b
bindS = flip concatMapS

-- Replacement for M.!
atDebug :: (Ord a, Show a) => M.Map a b -> a -> b
atDebug m a = case M.lookup a m of
                Nothing -> error $ "atDebug:  " ++ show a
                Just x -> x

-- * Model Checking CFG vs. NFA

-- | Auxiliary data.
-- Intuitively, if @(n,s)->S@ is in the table, the 'CFG' from @n@ generates a
-- string which makes the 'NFA' go from @s@ to any state in the set @S@
newtype CFGtable = CFGtable { table :: M.Map (NonTerm,State) (S.Set State) }
                deriving (Show, Eq, Ord)

-- | Computes the next approximation for 'CFGtable'.
-- Basically, try to run the 'NFA' in parallel with the 'CFG'.
nextReachables :: NFA -> CFG -> CFGtable -> CFGtable
nextReachables nfa cfg tab =
    let t0 = table tab
        prEv :: (Act,[Res]) -> Term
        prEv (a,r) = a ++ "(" ++ concat (intersperse "," r) ++ ")"
        reachNFA :: Term -> State -> S.Set State
        reachNFA term st = S.map snd $ S.filter ((==term) . prEv . fst) $
                           (trans nfa `atDebug` st)
        reachCFG :: NonTerm -> State -> S.Set State
        reachCFG nterm st = t0 `atDebug` (nterm, st)
        reachAny :: Symbol -> State -> S.Set State
        reachAny sym = if null sym || not (isUpper $ head sym)
                       then reachNFA sym else reachCFG sym
        -- We must not discard final states when the CFG has empty language!!
        addIfFinal :: State -> S.Set State -> S.Set State
        addIfFinal s ss = if s `S.member` final nfa then S.insert s ss else ss
        reach :: NonTerm -> State -> S.Set State
        reach n s = addIfFinal s
                    (prods cfg `atDebug` n
                     `bindS`
                     foldl' (flip (concatMapS . reachAny)) (S.singleton s))
        nextTable :: M.Map (NonTerm,State) (S.Set State)
                  -> (NonTerm, State)
                  -> M.Map (NonTerm,State) (S.Set State)
        nextTable t k@(nterm,s) = M.insertWith S.union k (reach nterm s) t
    in CFGtable $ foldl' nextTable t0 (M.keys t0)

-- |  PTIME Algorithm to check whether, given
-- a 'NFA' @A@ and a 'CFG' @G@ we have @L(A) \cap L(G) = \emptyset@
-- where @L@ means 'language of'. Note that @L(G)@ for our purposes
-- is defined so to ignore non-termination, e.g. @A -> a A@ has not
-- the empty language but rather comprises @a^n@ for all @n@.
emptyIntersection :: NFA -> CFG -> Bool
emptyIntersection nfa cfg =
    let states    = M.keys (trans nfa)
        nonTerms = M.keys (prods cfg)
        table0 = CFGtable $ M.fromList
                 [ ((n,s),S.empty) | n <- nonTerms , s <- states]
        reachableTable = fix (nextReachables nfa cfg) table0
        reachableStates =
            table reachableTable `atDebug` (starting cfg, initial nfa)
        in -- trace (show reachableStates) $
           S.null $ S.intersection reachableStates (final nfa)

-- * User interface

-- | The main driver
checkAll :: Either String FilePath -- ^ usage
         -> FilePath               -- ^ policy
         -> IO Bool
checkAll usage policyFile = do
  usageStr <- case usage of
                Left u -> return u
                Right usageFile -> readFile usageFile
  let u = case P.parse topLevelUsage "" usageStr of
            Left e -> error (show e)
            Right x -> x
  print u
  let reg_u = regularize u
  print reg_u

  policyStr <- readFile policyFile
  let (polName, pol) =  case P.parse policy policyFile policyStr of
            Left e -> error (show e)
            Right x -> x
  putStr "policy: "
  print polName
  print pol
  checkNFA pol
  let allActs = S.toList $ usageActs u `S.union` nfaActs pol
      inconsistAct = [ (act1,n1,n2)
                     | (act1,n1) <- allActs, (act2,n2) <- allActs
                     , act1==act2, n1/=n2 ]
  unless (null inconsistAct) $
    error $ "**** Error: Actions are used with inconsistent arity: " ++
       show inconsistAct

  let allPhis = S.toList $ usagePhis u `S.union` nfaPhis pol
                           `S.union` S.singleton polName
      phiActs = concat [ [('[':p,1), (']':p,1)] | p <- allPhis ]
  putStr "allActs: " >> print allActs
  putStr "allPhis: " >> print allPhis
  let occurNRes = usageRes u `S.union` nfaRes pol
      params = S.toList $ nfaPar pol
      sharps = [ '#':show n | n <- [1..length params] ]
      allRes = S.toList (S.fromList ("":"_":sharps)
                         `S.union` occurNRes)
  putStr "params: "
  print params
  putStr "known resources: "
  print allRes

  let bpa_u = makeBPA sharps reg_u
  print bpa_u

  let checkOne xMap = do
         putStrLn $ "*********** Checking " ++ show xMap
         let inst_pol = instantiate pol xMap allRes allActs
         -- print inst_pol
         let framed_pol = framed inst_pol polName allPhis
         putStrLn "*** Framed policy"
         -- print framed_pol
         checkNFA framed_pol
         putStrLn "*** WU policy"
         -- aSharp must be instatiated to add self loops!
         let aSh = aSharp (length params)
             aSh_inst = instantiate aSh [("x","#1")] allRes (allActs ++ phiActs)
         -- print aSh_inst
         checkNFA aSh_inst
         let wu_pol = framed_pol `weakUntil` aSh_inst
         -- print wu_pol
         checkNFA wu_pol
         putStrLn "*** Empty inters"
         let result = emptyIntersection wu_pol bpa_u
         let resStr | result    = "RESPECTS"
                    | otherwise = "VIOLATES"
         putStrLn $ "The usage " ++ resStr ++ " the policy."
         return result
  -- EXPTIME on the number of the params
  let allAssignments = map (zip params) $
                       sequence [ filter (/="_") allRes | _ <- params ]
  -- Stop after the first failure
  andM $ map checkOne allAssignments
  -- checkOne [("x1","#1"),("x2","#2")]

-- | Main
main :: IO ()
main = do
  args <- getArgs
  when (length args /= 2) $ error "usage: Main policyFile usage"
  let [policyFile, usage'] = args
      fileOpt = "--file="
      usage = if fileOpt `isPrefixOf` usage'
              then Right $ drop (length fileOpt) usage'
              else Left usage'
  res <- checkAll usage policyFile
  let result | res       = "PASS (no policy violation found)"
             | otherwise = "FAIL (some policy violations found)"
  putStrLn $ "***** Final result: " ++ result

-- * Tests

check0 :: IO Bool
check0 = checkAll (Right "USAGE_0") "NFA_0"

check1' :: IO Bool
check1' = checkAll (Right "USAGE_1") "NFA_1"

check1 :: String -> IO Bool
check1 u = checkAll (Left u) "POL_1"

check2, check3, check4, check5 :: IO Bool
check2 = check1 "phi[ nu n . new(n) ]"
check3 = check1 "phi[ nu n . new(n) . a(n) ]"
check4 = check1 "phi[ nu n . new(n) . a(n) . b(n) ]"
check5 = check1 "phi[ nu n . new(n) . a(n) . b(n) . (mu h . h) ]"
