module Parser where

import Data.List
import qualified Data.Set as S
import qualified Data.Map as M
import Control.Monad
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Char
import Common
import Policy as P
import Usage


-- * Usages

name :: Parser String
name = do
  n <- (lower <|> oneOf "-+*")
  ns <- many (alphaNum <|> oneOf "_-+*")
  spaces
  return (n:ns)
 <?> "name"

nilU :: Parser U
nilU = (string "0" >> spaces >> return Eps)

baseU :: Parser U
baseU =
    ((between (char '(' >> spaces) (char ')' >> spaces) usageU) <?> "(usage)")
     <|> nilU
     <|> nuU
     <|> muU
     <|> phiU
     <|> (do
         act_or_h <- name
         spaces
         (try (do char '(' >> spaces
                  res <- sepBy name (char ',' >> spaces)
                  let res' = if null res then [""] else res
                  char ')' >> spaces
                  return (Event act_or_h res')) )
          <|> return (Var act_or_h) <?> "recursion variable")

factU :: Parser U
factU = do
  u1 <- baseU
  (char '.' >> spaces >> liftM (Seq u1) factU) <|> return u1

termU :: Parser U
termU = do
  u1 <- factU
  (char '+' >> spaces >> liftM (Sum u1) termU) <|> return u1

nuU :: Parser U
nuU = do
  try (string "nu")
  spaces
  n <- many1 alphaNum
  spaces
  char '.' >> spaces
  u <- usageU
  return (Nu n u)
 <?> "nu"

muU :: Parser U
muU = do
  try (string "mu")
  spaces
  v <- name
  spaces
  char '.' >> spaces
  u <- usageU
  return (Rec v u)
 <?> "mu"

phiU :: Parser U
phiU = do
  p <- try (do p <- name
               spaces
               char '[' >> spaces
               return p)
  u <- usageU
  char ']' >> spaces
  return (Phi p u)
 <?> "framing"

usageU :: Parser U
usageU = nuU <|> muU <|> termU

topLevelUsage :: Parser U
topLevelUsage = do spaces ; u <- usageU ; eof ; return u

-- * Policies

initialP :: Parser P.State
initialP = string "start:" >> spaces >> name

nameExcept :: String -> Parser String
nameExcept exc = try (do n <- name ; if n == exc then mzero else return n)

finalP :: Parser [P.State]
finalP = string "final:" >> spaces >> many (nameExcept "trans")

transitionP :: Parser (M.Map P.State (S.Set ((Act, [Res]), P.State)))
transitionP = do
  s1 <- name
  ts <- many transP
  return $ M.singleton s1 (S.unions ts)

transP :: Parser (S.Set ((Act, [Res]), P.State))
transP = do
  string "--" >> spaces
  act <- name
  res <- (do char '(' >> spaces
             res <- sepBy name (char ',' >> spaces)
             let res' = if null res then [""] else res
             char ')' >> spaces
             return res'
          <|> return [""])
  string "-->" >> spaces
  s2 <- name
  return $ S.singleton ((act,res),s2)

statesP :: Parser [P.State]
statesP = do
  string "states:" >> spaces
  many1 (nameExcept "start")

policy :: Parser (String, NFA)
policy = do
  spaces
  string "name:" >> spaces
  n <- name
  st <- statesP
  let tr0 = M.fromList [ (s, S.empty) | s <- st ]
  i <- initialP
  f <- finalP
  string "trans:" >> spaces
  ts <- many transitionP
  let nfa = NFA { trans = foldl' (M.unionWith S.union) tr0 ts
                , initial = i
                , final = S.fromList f
                }
  return (n, nfa)
