{-# OPTIONS_GHC -fglasgow-exts #-}
module Policy where

import Common
import qualified Data.Map as M
import qualified Data.Set as S
import Data.Generics

type State = String
-- | Non-determinisitic finite state automaton (without epsilon transitions)
data NFA = NFA { trans :: M.Map State (S.Set ((Act, [Res]), State))
               , initial :: State
               , final :: S.Set State
               }
           deriving (Show, Read, Eq, Ord, Typeable, Data)

unionS :: Ord a => S.Set (S.Set a) -> S.Set a
unionS = S.unions . S.toList

-- | Instantiate @x@ with 'Res'
instantiate :: NFA
            -> [(Nam,Res)]   -- ^ how to replace @x@
            -> [Res]         -- ^ all the 'Res' we know
            -> [(Act,Int)]   -- ^ all the 'Act's with arity
            -> NFA
instantiate nfa r rs' acts = let
    rs = filter (not . flip elem (map snd r)) rs'
    tr :: M.Map State (S.Set ((Act,[Res]),State))
    tr = trans nfa
    instX, qmarks, selfLoops :: M.Map State (S.Set ((Act,[Res]),State))
    instX = M.map (unionS . S.map (\((act,res),st) ->
             let inst :: Res -> [Res]
                 inst r1 = case lookup r1 r of
                             Just aRes -> [aRes]
                             Nothing -> if r1=="-" then rs else [r1]
              in S.fromList
                [ ((act,res'),st) | res' <- sequence $ map inst res ]
            )) tr
    qmarks = M.map (unionS . S.map (\((act,res),st)->
              let resS = sequence [ [r1,"?"] | r1 <- res ]
              in S.fromList [ ((act,r1),st) | r1 <- resS ]
                          )) instX
    mUnion = M.unionWith S.union
    trBar = instX `mUnion` qmarks
    allEvents :: [(Act,[Res])]
    allEvents = [ (act,res)
                | (act, arity) <- acts
                , res <- sequence $ replicate arity ("?":rs') ]
    addSelfLoop :: State -> S.Set ((Act,[Res]),State)
                -> S.Set ((Act,[Res]),State)
    addSelfLoop state trSet =
        S.unions [ if any  ((ev'==) . fst) $ S.toList trSet
                   then S.empty else S.singleton (ev',state)
                 | ev' <- allEvents ]
    selfLoops = M.mapWithKey addSelfLoop trBar
    in NFA { trans = trBar `mUnion` selfLoops
           , initial = initial nfa
           , final = final nfa
           }

-- | Construct the two-layered automaton that understands @[phi@ and @]phi@
framed :: NFA      -- ^ policy automaton
       -> String   -- ^ policy name
       -> [String] -- ^ other policies (so we skip related events)
       -> NFA      -- ^ framed version
framed nfa phi phis' = let
    phis = filter (/=phi) phis'
    patchSt s = S.map (\(str,st) -> (str,s++st))
    patchStLoop stL = S.map (\(str,_st) -> (str,stL))
    tr = trans nfa
    tr1 = M.mapKeysMonotonic ("up_"++) $ M.map (patchSt "up_") tr
    tr2 = M.mapKeysMonotonic ("dn_"++) $
          M.mapWithKey (\s e -> if s `S.member` final nfa
                                then patchStLoop ("dn_"++s) e
                                else patchSt "dn_" e) tr
    states = M.keys (trans nfa)
    states1 = map ("up_"++) states
    states2 = map ("dn_"++) states
    states2NF = map ("dn_"++) $ filter (not . (`S.member` final nfa)) states
    links = M.fromList $ [ (s1, S.singleton (("["++phi,[""]), s2))
                               | (s1,s2) <- zip states1 states2 ] ++
                         [ (s2, S.singleton (("]"++phi,[""]), s1))
                               | (s1',s2) <- zip states1 states2
                               , let s1 = if s2 `elem` states2NF
                                          then s1' else s2
                         ]
    selfLoops = M.fromList [ (s, S.fromList [(("["++p,[""]),s),
                                             (("]"++p,[""]),s)])
                                 | s <- states1 ++ states2NF , p <- phis]
    -- Note: M.union does not merge sets, so we must be careful here
    trUnion = M.unionWith S.union
    in NFA { trans = tr1 `trUnion` tr2 `trUnion` links `trUnion` selfLoops
           , initial = "up_" ++ initial nfa
           , final = S.map ("dn_"++) $ final nfa
           }

-- * A# automaton

-- | This is a policy, to be instantiated to add self loops
aSharp :: Int -- ^ the number of # witnesses
       -> NFA
aSharp n = let
    st :: Int -> String
    st s = "s" ++ show s
    tr :: Int -> S.Set ((Act,[Res]),State)
    tr s = S.fromList $
            [ (("new", ["#"++show i]), "sink") | i <- [1..n], i/=s ] ++
            [ (("new", ["#"++show s]), st (s+1)) ]
    trLast = S.fromList [ (("new", ["#"++show n]), "sink") ]
    in NFA { trans = M.fromList $
              [ (st s, tr s) | s <- [1..n] ] ++ [ (st (n+1), trLast) ] ++
              [ ("sink", S.empty) ]
           , initial = "s1"
           , final = S.singleton "sink"
           }

-- * Weak until

weakUntil :: NFA -> NFA -> NFA
weakUntil nfa0 nfa1 = let
    escape :: String -> String
    escape [] = []
    escape ('-':rest) = '-':'-':escape rest
    escape (c:rest) = c:escape rest
    pair :: String -> String -> String
    pair x y = escape x ++ ">-<" ++ escape y
    {-
    unescape :: String -> String
    unescape [] = []
    unescape ('-':'-':rest) = '-':unescape rest
    unescape (c:rest) = c:unescape rest
    unpair :: String -> (String,String)
    unpair s = let unpair' [] = error "unpair'"
                   unpair' ('>':'-':'<':rest) = ("",rest)
                   unpair' (c:rest) = let (pre',post') = unpair' rest
                                      in (c:pre',post')
                   (pre,post) = unpair' s
                   in (unescape pre, unescape post)
     -}
    states0 = M.keys (trans nfa0)
    states0F = filter (`S.member` final nfa0) states0
    states1NF = S.toList (M.keysSet (trans nfa1) `S.difference` final nfa1)
    thetas = nubOrd [ th | (th, _) <- S.toList $ S.unions $
                (M.elems $ trans nfa0) ++ (M.elems $ trans nfa1) ]
    fin = [ pair q0 q1 | q0 <- states0F , q1 <- states1NF ]
    in NFA { trans = M.fromListWith S.union $
                     [ (pair q0 q1, S.singleton (th, pair q0' q1'))
                     | q0        <- states0
                     , (th, q0') <- S.toList (trans nfa0 M.! q0)
                     , q1        <- states1NF
                     , (th', q1')<- S.toList (trans nfa1 M.! q1)
                     , th == th' ] ++
                     [ (pair q0 q1, S.singleton (th, pair q0 q1))
                     | q0 <- states0
                     , q1 <- S.toList $ final nfa1
                     , th <- thetas
                     ] ++
                     [ (qF, S.empty) | qF <- fin ]
           , initial = pair (initial nfa0) (initial nfa1)
           , final = S.fromList fin
           }

-- * 'Res','Act',etc. occurring in a 'NFA'

isParam :: NRes -> Bool
isParam "-"     = True
isParam ('x':_) = True
isParam _       = False

nfaPhis :: NFA -> S.Set String
nfaPhis = let
    getPhi :: (Act, [Res]) -> S.Set String
    getPhi ([],_res) = S.empty
    getPhi (act,_res) = if last act == '['
                        then S.singleton $ init act
                        else S.empty
    in everything (S.union) (mkQ S.empty getPhi)

-- | 'Act's and their arity
nfaActs :: NFA -> S.Set (Act, Int)
nfaActs = let
    getAct :: (Act, [NRes]) -> S.Set (Act, Int)
    getAct (act,res) = S.singleton (act, length res)
    in everything (S.union) (mkQ S.empty getAct)

nfaRes :: NFA -> S.Set Res
nfaRes = let
    getRes :: (Act, [NRes]) -> S.Set Res
    getRes (_act,res) = S.fromList $ filter (not . isParam) res
    in (`S.difference` S.singleton "-") .
       everything (S.union) (mkQ S.empty getRes)

nfaPar :: NFA -> S.Set String
nfaPar = let
    getPar :: (Act, [NRes]) -> S.Set String
    getPar (_act,res) = S.fromList $ filter isParam res
    in everything (S.union) (mkQ S.empty getPar)

checkNFA :: NFA -> IO ()
checkNFA nfa = let
    states, states' :: S.Set State
    states = M.keysSet (trans nfa)
    states'= S.map snd $ S.unions $ M.elems $ trans nfa
    diff = states' `S.difference` states
    in if S.null diff then do
             -- putStrLn "NFA ok"
             -- print states
             -- print states'
             return ()
       else do
             putStrLn "NFA NOT ok !!!!"
             print states
             print states'
             print diff
