{-# OPTIONS_GHC -fglasgow-exts #-}
{- -- # LANGUAGE DeriveDataTypeable #-}
module Usage where

import Common
import CFG
import Control.Monad.State hiding (fix)
import qualified Data.Map as M
import qualified Data.Set as S
import Data.Generics
import Data.List
-- import Debug.Trace

data U
  = Eps
  | Event Act [NRes]
  | Seq U U
  | Sum U U
  | Nu Nam U
  | Phi String U
  | Var Rvar
  | Rec Rvar U
  deriving (Eq, Ord, Typeable, Data)

instance Show U where
    showsPrec _ Eps = ("0"++)
    showsPrec _ (Event act nres) =
        \s -> act ++ "(" ++ concat (intersperse "," nres) ++ ")" ++ s
    showsPrec n (Sum u1 u2) = showParen (n>1) $
        showsPrec 1 u1 . (" + "++) . showsPrec 1 u2
    showsPrec _ (Seq u1 u2) = showsPrec 2 u1 . ("."++) . showsPrec 2 u2
    showsPrec _ (Phi p u) = (p++) . ("["++) . showsPrec 0 u . ("]"++)
    showsPrec _ (Var rvar) = (rvar++)
    showsPrec n (Nu n' u) = showParen (n>0) $
                ("nu "++) . (n' ++)  . (". "++) . showsPrec 0 u
    showsPrec n (Rec rvar u) = showParen (n>0) $
                ("mu "++) . (rvar++) . (". "++) . showsPrec 0 u

-- * Regularization

-- | Remove the @phi[]@ framings in a 'U'
removePhi :: String -> U -> U
removePhi _   Eps = Eps
removePhi _   u@(Event _ _) = u
removePhi phi u'@(Phi phi' u) = if phi==phi' then u else u'
removePhi phi u = gmapT (mkT $ removePhi phi) u

-- | Find a fresh 'Rvar' for a 'U'
freshRvar :: Rvar -- ^ hint: return this, if possible
          -> U
          -> U
          -> Rvar
freshRvar v u uCh =
    let freshIn v' (Var v'') = v' /= v''
        freshIn v' (Rec v'' u1) = v' == v'' || freshIn v' u1
        freshIn v' u1 = gmapQl (&&) True (mkQ True (freshIn v')) u1
        genFresh v' = if v' `freshIn` (u `Seq` uCh) then v' else v' ++ "$"
        in if v `freshIn` u then v else fix genFresh (v++"$")

-- | Find a fresh 'Nam' for a 'U'
freshNam :: Nam -- ^ hint: return this, if possible
         -> U   -- ^ must be fresh here
         -> U   -- ^ if changed, must also be fresh here
         -> Nam
freshNam n u uCh = let
        -- freshIn n' u | trace (show n'++ " "++show u) False = undefined
        freshIn n' (Event _ nres) = not $ n' `elem` nres
        freshIn n' (Nu n'' u1) = n' == n'' || freshIn n' u1
        freshIn n' u1 = gmapQl (&&) True (mkQ True (freshIn n')) u1
        genFresh n' = if n' `freshIn` (u `Seq` uCh) then n' else n' ++ "$"
        in if n `freshIn` u then n else fix genFresh (n++"$")

alphaConvertNu :: U -- ^ the replacement
               -> U -- ^ the 'Nu'
               -> U
alphaConvertNu replU (Nu n u) = let n' = freshNam n replU u
                                in Nu n' (substU n n' u)
alphaConvertNu _     _ = error "alphaConvertNu"

alphaConvertMu :: U -- ^ the replacement
               -> U -- ^ the 'Mu'
               -> U
alphaConvertMu replU (Rec v u) = let v' = freshRvar v replU u
                                 in Rec v' (substV v (Var v') u)
alphaConvertMu _     _ = error "alphaConvertMu"

-- | Replace a 'Nam' with another 'Nam' or 'Res'
substU :: Nam -> NRes -> U -> U
substU n r (Event act nres) =
    Event act $ map (\nr -> if nr == n then r else nr) nres
substU n r u@(Nu n1 _) = if n == n1 then u
                         else let (Nu n2 u2) = alphaConvertNu (Event "" [r]) u
                              in Nu n2 (substU n r u2)
substU n r u = gmapT (mkT $ substU n r) u

-- | Replace a 'Rvar' with a 'U'
substV :: Nam -> U -> U -> U
substV v uR u@(Var v') = if v' == v then uR else u
substV v uR u@(Nu _ _) = let (Nu n' u') = alphaConvertNu uR u
                         in Nu n' (substV v uR u')
substV v uR u@(Rec v1 _) = if v == v1 then u
                           else let (Rec v2 u2) = alphaConvertMu uR u
                                in Rec v2 (substV v uR u2)
substV v uR u = gmapT (mkT $ substV v uR) u

-- | Substitute the occurrences of a Rvar under a phi, removing inner phi's.
-- (Regularization of @mu h. h@)
substRec :: Rvar     -- ^ var to replace
         -> U        -- ^ replacement
         -> [String] -- ^ active phi's
         -> U        -- ^ where to replace it
         -> U
substRec rv u p u'@(Nu _ _) = let (Nu n2 u2) = alphaConvertNu u u'
                                 in Nu n2 (substRec rv u p u2)
substRec rv u p u'@(Rec rv1 _) = if rv == rv1 then u'
                                    else let Rec rv2 u2 = alphaConvertMu u u'
                                         in Rec rv2 (substRec rv u p u2)
substRec rv u p (Phi phi u1) = Phi phi (substRec rv u (phi:p) u1)
substRec rv u p u'@(Var rv') = if rv /= rv' || null p then u'
                               else foldr removePhi u p
substRec rv u p u' = gmapT (mkT $ substRec rv u p) u'

-- | Remove redundant framings from a 'U'
regularize :: U -> U
regularize (Phi phi u) = Phi phi $ regularize (removePhi phi u)
regularize u@(Rec rvar u') = Rec rvar $ substRec rvar u [] (regularize u')
regularize u = gmapT (mkT $ regularize) u

-- * Transformation into BPA

type Supply a = State Int a

newNonTerm :: Supply NonTerm
newNonTerm = do n <- get ; modify succ ; return ("N_" ++ show n)

newNonTermRec :: Supply NonTerm
newNonTermRec = do n <- get ; modify succ ; return ("R_" ++ show n)

mUnion :: (Ord a, Ord b)
          => M.Map a (S.Set b) -> M.Map a (S.Set b) -> M.Map a (S.Set b)
mUnion = M.unionWith S.union

-- | Join the languages of two 'CFG's
sumCFG :: CFG -> CFG -> Supply CFG
sumCFG g1 g2 = do
  n <- newNonTerm
  return CFG { prods = M.singleton n (S.singleton [starting g1])
                  `mUnion` M.singleton n (S.singleton [starting g2])
                  `mUnion` prods g1 `mUnion` prods g2
             , starting = n }

type Env = Rvar -> NonTerm

env0 :: Env
env0 = \rvar -> error ("Rvar " ++ rvar ++ " not defined")

updateEnv :: Rvar -> NonTerm -> Env -> Env
updateEnv rvar nt r = \rvar' -> if rvar==rvar' then nt else r rvar'

-- | Extract a BPA ('CFG') from a 'U'
bpa :: [Res]    -- ^ available witnesses @#@
    -> Env      -- ^ recursion variables
    -> U
    -> Supply CFG
bpa _ _ Eps = do
  nt <- newNonTerm
  return CFG { prods = M.singleton nt (S.singleton [])
             , starting = nt }
bpa _ _ (Event act nres) = do
  nt <- newNonTerm
  return CFG { prods = M.singleton nt (S.singleton
                                       [act ++ "(" ++
                                        concat (intersperse "," nres) ++
                                        ")"])
             , starting = nt }
bpa n r (Seq u1 u2) = do
  g1 <- bpa n r u1
  g2 <- bpa n r u2
  nt <- newNonTerm
  return CFG { prods = M.singleton nt (S.singleton [starting g1, starting g2])
                   `mUnion` prods g1 `mUnion` prods g2
             , starting = nt }
bpa n r (Sum u1 u2) = do
  g1 <- bpa n r u1
  g2 <- bpa n r u2
  sumCFG g1 g2
bpa [] r (Nu nam u) = bpa [] r (substU nam "_" u)
bpa n@(w:n') r (Nu nam u) = do
  g1 <- bpa n  r (substU nam "_" u)
  g2 <- bpa n' r (substU nam w u)
  sumCFG g1 g2
bpa n r (Phi phi u) =
  bpa n r (Event ("["++phi) [""] `Seq` u `Seq` Event ("]"++phi) [""])
bpa _ r (Var rvar) = return CFG { prods = M.empty , starting = r rvar }
bpa n r (Rec rvar u) = do
 nt <- newNonTermRec
 b <- bpa n (updateEnv rvar nt r) u
 return $ b { prods = prods b
                      `mUnion` M.singleton nt (S.singleton [starting b]) }

makeBPA :: [Res] -> U -> CFG
makeBPA shs u = evalState (bpa shs env0 u) 0

usagePhis :: U -> S.Set String
usagePhis = let
    getPhi :: U -> S.Set String
    getPhi (Phi phi _) = S.singleton phi
    getPhi _ = S.empty
    in everything S.union (mkQ S.empty getPhi)

-- | 'Act's and their arity
usageActs :: U -> S.Set (Act, Int)
usageActs = let
    getAct :: U -> S.Set (Act, Int)
    getAct (Event act nres) = S.singleton (act, length nres)
    getAct _ = S.empty
    in everything S.union (mkQ S.empty getAct)

usageRes :: U -> S.Set Res
usageRes (Event _ res) = S.fromList res
usageRes (Nu n u) = usageRes u `S.difference` S.singleton n
usageRes u = gmapQl S.union S.empty (S.empty `mkQ` usageRes) u

-- tests

test0 :: U
test0 = Event "a" ["r"]
        `Seq`
        Rec "h" (Nu "n" (Event "b" ["n"]
                         `Seq`
                         Event "b" ["n"]
                         `Seq`
                         Var "h"
                         `Seq`
                         Event "b" ["n"]))

test1 :: U
test1 = Rec "h" (Event "b" ["n"]
                 `Seq`
                 Var "h"
                 `Seq`
                 Phi "phi_cw" (Var "h")
                 `Seq`
                 Event "b" ["n"])

test2 :: U
test2 = Rec "h1" (Var "h1"
                 `Seq`
                 Phi "phi_cw" (Var "h1")
                 `Seq`
                 (Rec "h2"
                      (Var "h1"
                       `Seq`
                       Phi "phi2" (Var "h2")
                      )))
